@@ -1,4 +1,4 @@
-# Copyright (C) 2017 Open Information Security Foundation
+# Copyright (C) 2017-2022 Open Information Security Foundation
 # Copyright (c) 2015-2017 Jason Ish
 #
 # You can copy, redistribute or modify this Program under the terms of
@@ -99,19 +99,26 @@ INDEX_EXPIRATION_TIME = 60 * 60 * 24 * 14
 # Rule keywords that come with files
 file_kw = ["filemd5", "filesha1", "filesha256", "dataset"]
 
+def strict_error(msg):
+    logger.error(msg)
+    if config.args().fail:
+        sys.exit(1)
+
 class Fetch:
 
     def __init__(self):
         self.istty = os.isatty(sys.stdout.fileno())
 
-    def check_checksum(self, tmp_filename, url):
+    def check_checksum(self, tmp_filename, url, checksum_url=None):
         try:
-            checksum_url = url + ".md5"
+            if not isinstance(checksum_url, str):
+                checksum_url = url[0] + ".md5"
+            net_arg=(checksum_url,url[1])
             local_checksum = hashlib.md5(
                 open(tmp_filename, "rb").read()).hexdigest().strip()
             remote_checksum_buf = io.BytesIO()
             logger.info("Checking %s." % (checksum_url))
-            net.get(checksum_url, remote_checksum_buf)
+            net.get(net_arg, remote_checksum_buf)
             remote_checksum = remote_checksum_buf.getvalue().decode().strip()
             logger.debug("Local checksum=|%s|; remote checksum=|%s|" % (
                 local_checksum, remote_checksum))
@@ -163,7 +170,7 @@ class Fetch:
             logger.info("Using latest cached version of rule file: %s", url)
             if not os.path.exists(tmp_filename):
                 logger.error("Can't proceed offline, "
-                             "source %s has not yet been downloaded.", url)
+                             "source {} has not yet been downloaded.".format(url))
                 sys.exit(1)
             return self.extract_files(tmp_filename)
         if not config.args().force and os.path.exists(tmp_filename):
@@ -174,7 +181,7 @@ class Fetch:
                     url)
                 return self.extract_files(tmp_filename)
             if checksum:
-                if self.check_checksum(tmp_filename, url):
+                if self.check_checksum(tmp_filename, net_arg, checksum):
                     logger.info("Remote checksum has not changed. "
                                 "Not fetching.")
                     return self.extract_files(tmp_filename)
@@ -191,14 +198,15 @@ class Fetch:
             tmp_fileobj.close()
         except URLError as err:
             if os.path.exists(tmp_filename):
-                logger.warning(
-                    "Failed to fetch %s, "
-                    "will use latest cached version: %s", url, err)
+                if config.args().fail:
+                    strict_error("Failed to fetch {}: {}".format(url, err))
+                else:
+                    logger.error("Failed to fetch {}, will use latest cached version: {}".format(url, err))
                 return self.extract_files(tmp_filename)
             raise err
         except IOError as err:
             self.progress_hook_finish()
-            logger.error("Failed to copy file: %s", err)
+            logger.error("Failed to copy file: {}".format(err))
             sys.exit(1)
         except Exception as err:
             raise err
@@ -214,7 +222,7 @@ class Fetch:
                 files.update(fetched)
             except URLError as err:
                 url = url[0] if isinstance(url, tuple) else url
-                logger.error("Failed to fetch %s: %s", url, err)
+                strict_error("Failed to fetch {}: {}".format(url, err))
         else:
             for url in self.args.url:
                 files.update(self.fetch(url))
@@ -240,12 +248,16 @@ def load_filters(filename):
             line = line.strip()
             if not line or line.startswith("#"):
                 continue
-            line = line.rsplit(" #")[0]
+            line = line.rsplit(" #")[0].strip()
 
-            line = re.sub(r'\\\$', '$', line)  # needed to escape $ in pp
             try:
-                rule_filter = matchers_mod.ModifyRuleFilter.parse(line)
-                filters.append(rule_filter)
+                if line.startswith("metadata-add"):
+                    rule_filter = matchers_mod.AddMetadataFilter.parse(line)
+                    filters.append(rule_filter)
+                else:
+                    line = re.sub(r'\\\$', '$', line)  # needed to escape $ in pp
+                    rule_filter = matchers_mod.ModifyRuleFilter.parse(line)
+                    filters.append(rule_filter)
             except Exception as err:
                 raise exceptions.ApplicationError(
                     "Failed to parse modify filter: {}".format(line))
@@ -295,6 +307,7 @@ def load_local(local, files):
         if len(local_files) == 0:
             local_files.append(local)
         for filename in local_files:
+            filename = os.path.realpath(filename)
             logger.info("Loading local file %s" % (filename))
             if filename in files:
                 logger.warn(
@@ -304,7 +317,7 @@ def load_local(local, files):
                 with open(filename, "rb") as fileobj:
                     files.append(SourceFile(filename, fileobj.read()))
             except Exception as err:
-                logger.error("Failed to open %s: %s" % (filename, err))
+                logger.error("Failed to open {}: {}".format(filename, err))
 
 def load_dist_rules(files):
     """Load the rule files provided by the Suricata distribution."""
@@ -359,7 +372,7 @@ def load_dist_rules(files):
                 with open(path, "rb") as fileobj:
                     files.append(SourceFile(path, fileobj.read()))
             except Exception as err:
-                logger.error("Failed to open %s: %s" % (path, err))
+                logger.error("Failed to open {}: {}".format(path, err))
                 sys.exit(1)
 
 def load_classification(suriconf, files):
@@ -422,15 +435,48 @@ def manage_classification(suriconf, files):
 def handle_dataset_files(rule, dep_files):
     if not rule.enabled:
         return
-    load_attr = [el.strip() for el in rule.dataset.split(",") if "load" in el][0]
-    dataset_fname = os.path.basename(load_attr.split(" ")[1])
-    filename = [fname for fname, content in dep_files.items() if fname == dataset_fname]
-    if filename:
-        logger.debug("Copying dataset file %s to output directory" % dataset_fname)
-        with open(os.path.join(config.get_output_dir(), dataset_fname), "w+") as fp:
-            fp.write(dep_files[dataset_fname].decode("utf-8"))
+
+    dataset_load = [el.strip() for el in rule.dataset.split(",") if el.startswith("load")]
+    if not dataset_load:
+        # No dataset load found.
+        return
+    dataset_filename = dataset_load[0].split(maxsplit=1)[1].strip()
+
+    # Get the directory name the rule is from.
+    prefix = os.path.dirname(rule.group)
+
+    # Construct the source filename.
+    source_filename = "{}/{}".format(prefix, dataset_filename)
+
+    # If a source filename starts with a "/", look for it on the filesystem. The archive
+    # unpackers will take care of removing a leading / so this shouldn't happen for
+    # downloaded rulesets.
+    if source_filename.startswith("/"):
+        if not os.path.exists(source_filename):
+            logger.warn("Local dataset file '{}' was not found for rule {}, rule will be disabled".format(source_filename, rule.idstr))
+            rule.enabled = False
+            return
+        dataset_contents = open(source_filename, "rb").read()
     else:
-        logger.error("Dataset file %s was not found" % dataset_fname)
+        if not source_filename in dep_files:
+            logger.warn("Dataset file '{}' was not found for rule {}, rule will be disabled".format(dataset_filename, rule.idstr))
+            rule.enabled = False
+            return
+        dataset_contents = dep_files[source_filename]
+
+    content_hash = hashlib.md5(dataset_contents).hexdigest()
+    new_rule = re.sub("(dataset.*?load\s+){}".format(dataset_filename), "\g<1>datasets/{}".format(content_hash), rule.format())
+    dest_filename = os.path.join(config.get_output_dir(), "datasets", content_hash)
+    dest_dir = os.path.dirname(dest_filename)
+    logger.debug("Copying dataset file {} to {}".format(dataset_filename, dest_filename))
+    try:
+        os.makedirs(dest_dir, exist_ok=True)
+    except Exception as err:
+        logger.error("Failed to create directory {}: {}".format(dest_dir, err))
+        return
+    with open(dest_filename, "w") as fp:
+        fp.write(dataset_contents.decode("utf-8"))
+    return new_rule
 
 def handle_filehash_files(rule, dep_files, fhash):
     if not rule.enabled:
@@ -451,7 +497,7 @@ def handle_filehash_files(rule, dep_files, fhash):
         with open(os.path.join(filepath, os.path.basename(filehash_fname)), "w+") as fp:
             fp.write(dep_files[os.path.join("rules", filehash_fname)].decode("utf-8"))
     else:
-        logger.error("%s file %s was not found" % (fhash, filehash_fname))
+        logger.error("{} file {} was not found".format(fhash, filehash_fname))
 
 def write_merged(filename, rulemap, dep_files):
 
@@ -489,13 +535,17 @@ def write_merged(filename, rulemap, dep_files):
     with io.open(tmp_filename, encoding="utf-8", mode="w") as fileobj:
         for sid in rulemap:
             rule = rulemap[sid]
+            reformatted = None
             for kw in file_kw:
                 if kw in rule:
                     if "dataset" == kw:
-                        handle_dataset_files(rule, dep_files)
+                        reformatted = handle_dataset_files(rule, dep_files)
                     else:
                         handle_filehash_files(rule, dep_files, kw)
-            print(rule.format(), file=fileobj)
+            if reformatted:
+                print(reformatted, file=fileobj)
+            else:
+                print(rule.format(), file=fileobj)
     os.rename(tmp_filename, filename)
 
 def write_to_directory(directory, files, rulemap, dep_files):
@@ -543,22 +593,20 @@ def write_to_directory(directory, files, rulemap, dep_files):
             content = []
             for line in io.StringIO(file.content.decode("utf-8")):
                 rule = rule_mod.parse(line)
-                if not rule:
+                if not rule or rule.id not in rulemap:
                     content.append(line.strip())
                 else:
+                    reformatted = None
                     for kw in file_kw:
                         if kw in rule:
                             if "dataset" == kw:
-                                handle_dataset_files(rule, dep_files)
+                                reformatted = handle_dataset_files(rulemap[rule.id], dep_files)
                             else:
-                                handle_filehash_files(rule, dep_files, kw)
-                    if rule.id in rulemap:
-                        content.append(rulemap[rule.id].format())
+                                handle_filehash_files(rulemap[rule.id], dep_files, kw)
+                    if reformatted:
+                        content.append(reformatted)
                     else:
-                        # Just pass the input through. Most likey a
-                        # rule from a file that was ignored, but we'll
-                        # still pass it through.
-                        content.append(line.strip())
+                        content.append(rulemap[rule.id].format())
             tmp_filename = ".".join([outpath, "tmp"])
             io.open(tmp_filename, encoding="utf-8", mode="w").write(
                 u"\n".join(content))
@@ -891,6 +939,9 @@ def load_sources(suricata_version):
                         "Source index is required for source %s; "
                         "run suricata-update update-sources" % (source["source"]))
                 source_config = index.get_source_by_name(name)
+                if source_config is None:
+                    logger.warn("Source no longer exists in index and will not be fetched: {}".format(name))
+                    continue
                 try:
                     checksum = source_config["checksum"]
                 except:
@@ -1023,7 +1074,7 @@ def _main():
     ]
     for arg in unimplemented_args:
         if hasattr(args, arg) and getattr(args, arg):
-            logger.error("--%s not implemented", arg)
+            logger.error("--{} not implemented".format(arg))
             return 1
 
     suricata_path = config.get("suricata")
@@ -1034,7 +1085,7 @@ def _main():
         # The Suricata version was passed on the command line, parse it.
         suricata_version = engine.parse_version(args.suricata_version)
         if not suricata_version:
-            logger.error("Failed to parse provided Suricata version: %s" % (
+            logger.error("Failed to parse provided Suricata version: {}".format(
                 args.suricata_version))
             return 1
         logger.info("Forcing Suricata version to %s." % (suricata_version.full))
@@ -1061,7 +1112,7 @@ def _main():
         elif hasattr(args, "func"):
             return args.func()
         elif args.subcommand != "update":
-            logger.error("Unknown command: %s", args.subcommand)
+            logger.error("Unknown command: {}".format(args.subcommand))
             return 1
 
     if args.dump_sample_configs:
@@ -1177,6 +1228,12 @@ def _main():
 
     for key, rule in rulemap.items():
 
+        # To avoid duplicate counts when a rule has more than one modification
+        # to it, we track the actions here then update the counts at the end.
+        enabled = False
+        modified = False
+        dropped = False
+
         for matcher in disable_matchers:
             if rule.enabled and matcher.match(rule):
                 logger.debug("Disabling: %s" % (rule.brief()))
@@ -1187,25 +1244,32 @@ def _main():
             if not rule.enabled and matcher.match(rule):
                 logger.debug("Enabling: %s" % (rule.brief()))
                 rule.enabled = True
-                enable_count += 1
+                enabled = True
 
         for fltr in drop_filters:
             if fltr.match(rule):
-                rulemap[rule.id] = fltr.run(rule)
-                drop_count += 1
+                rule = fltr.run(rule)
+                dropped = True
 
         for fltr in modify_filters:
             if fltr.match(rule):
-                new_rule = fltr.run(rule)
-                if new_rule:
-                    rulemap[rule.id] = new_rule
-                    modify_count += 1
+                rule = fltr.run(rule)
+                modified = True
+
+        if enabled:
+            enable_count += 1
+        if modified:
+            modify_count += 1
+        if dropped:
+            drop_count += 1
+
+        rulemap[key] = rule
 
     # Check if we should disable ja3 rules.
     try:
         disable_ja3(suriconf, rulemap, disabled_rules)
     except Exception as err:
-        logger.error("Failed to dynamically disable ja3 rules: %s" % (err))
+        logger.error("Failed to dynamically disable ja3 rules: {}".format(err))
 
     # Check rule vars, disabling rules that use unknown vars.
     check_vars(suriconf, rulemap)
@@ -1224,7 +1288,7 @@ def _main():
     # Check that output directory is writable.
     if not os.access(config.get_output_dir(), os.W_OK):
         logger.error(
-            "Output directory is not writable: %s", config.get_output_dir())
+            "Output directory is not writable: {}".format(config.get_output_dir()))
         return 1
 
     # Backup the output directory.
@@ -1287,7 +1351,7 @@ def _main():
         logger.info("Running %s." % (config.get("reload-command")))
         rc = subprocess.Popen(config.get("reload-command"), shell=True).wait()
         if rc != 0:
-            logger.error("Reload command exited with error: %d", rc)
+            logger.error("Reload command exited with error: {}".format(rc))
 
     logger.info("Done.")
 
